import _extends from "@babel/runtime/helpers/esm/extends";
import { CancelledError, Console, functionalUpdate, getStatusProps, isCancelable, isCancelledError, isDocumentVisible, isOnline, isServer, isValidTimeout, noop, replaceEqualDeep, sleep } from './utils';
import { QueryStatus } from './types';
import { QueryObserver } from './queryObserver'; // TYPES

function _empty() {}

var ActionType = {
  Failed: 0,
  Fetch: 1,
  Success: 2,
  Error: 3
};

// CLASS
function _awaitIgnored(value, direct) {
  if (!direct) {
    return value && value.then ? value.then(_empty) : Promise.resolve();
  }
}

function _invokeIgnored(body) {
  var result = body();

  if (result && result.then) {
    return result.then(_empty);
  }
}

function _await(value, then, direct) {
  if (direct) {
    return then ? then(value) : value;
  }

  if (!value || !value.then) {
    value = Promise.resolve(value);
  }

  return then ? value.then(then) : value;
}

function _catch(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }

  if (result && result.then) {
    return result.then(void 0, recover);
  }

  return result;
}

function _invoke(body, then) {
  var result = body();

  if (result && result.then) {
    return result.then(then);
  }

  return then(result);
}

function _async(f) {
  return function () {
    for (var args = [], i = 0; i < arguments.length; i++) {
      args[i] = arguments[i];
    }

    try {
      return Promise.resolve(f.apply(this, args));
    } catch (e) {
      return Promise.reject(e);
    }
  };
}

export var Query = /*#__PURE__*/function () {
  function Query(config) {
    this.config = config;
    this.queryKey = config.queryKey;
    this.queryHash = config.queryHash;
    this.queryCache = config.queryCache;
    this.cacheTime = config.cacheTime;
    this.observers = [];
    this.state = getDefaultState(config);
    this.scheduleGc();
  }

  var _proto = Query.prototype;

  _proto.updateConfig = function updateConfig(config) {
    this.config = config;
    this.cacheTime = Math.max(this.cacheTime, config.cacheTime);
  };

  _proto.dispatch = function dispatch(action) {
    this.state = queryReducer(this.state, action);
    this.observers.forEach(function (observer) {
      observer.onQueryUpdate(action);
    });
    this.queryCache.notifyGlobalListeners(this);
  };

  _proto.scheduleGc = function scheduleGc() {
    var _this = this;

    if (isServer) {
      return;
    }

    this.clearGcTimeout();

    if (this.observers.length > 0 || !isValidTimeout(this.cacheTime)) {
      return;
    }

    this.gcTimeout = setTimeout(function () {
      _this.clear();
    }, this.cacheTime);
  };

  _proto.cancel = function cancel(silent) {
    try {
      var _this3 = this;

      var promise = _this3.promise;
      return _invokeIgnored(function () {
        if (promise && _this3.cancelFetch) {
          _this3.cancelFetch(silent);

          return _awaitIgnored(promise.catch(noop));
        }
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.continue = function _continue() {
    var _this$continueFetch;

    (_this$continueFetch = this.continueFetch) == null ? void 0 : _this$continueFetch.call(this);
  };

  _proto.clearTimersObservers = function clearTimersObservers() {
    this.observers.forEach(function (observer) {
      observer.clearTimers();
    });
  };

  _proto.clearGcTimeout = function clearGcTimeout() {
    if (this.gcTimeout) {
      clearTimeout(this.gcTimeout);
      this.gcTimeout = undefined;
    }
  };

  _proto.setData = function setData(updater, options) {
    var _this$config$isDataEq, _this$config;

    var prevData = this.state.data; // Get the new data

    var data = functionalUpdate(updater, prevData); // Structurally share data between prev and new data if needed

    if (this.config.structuralSharing) {
      data = replaceEqualDeep(prevData, data);
    } // Use prev data if an isDataEqual function is defined and returns `true`


    if ((_this$config$isDataEq = (_this$config = this.config).isDataEqual) == null ? void 0 : _this$config$isDataEq.call(_this$config, prevData, data)) {
      data = prevData;
    } // Try to determine if more data can be fetched


    var canFetchMore = hasMorePages(this.config, data); // Set data and mark it as cached

    this.dispatch({
      type: ActionType.Success,
      data: data,
      canFetchMore: canFetchMore,
      updatedAt: options == null ? void 0 : options.updatedAt
    });
  };

  _proto.clear = function clear() {
    this.queryCache.removeQuery(this);
  };

  _proto.destroy = function destroy() {
    this.clearGcTimeout();
    this.clearTimersObservers();
    this.cancel();
  };

  _proto.isEnabled = function isEnabled() {
    return this.observers.some(function (observer) {
      return observer.config.enabled;
    });
  };

  _proto.isStale = function isStale() {
    return this.observers.some(function (observer) {
      return observer.getCurrentResult().isStale;
    });
  };

  _proto.isStaleByTime = function isStaleByTime(staleTime) {
    if (staleTime === void 0) {
      staleTime = 0;
    }

    return !this.state.isSuccess || this.state.updatedAt + staleTime <= Date.now();
  };

  _proto.onInteraction = function onInteraction(type) {
    // Execute the first observer which is enabled,
    // stale and wants to refetch on this interaction.
    var staleObserver = this.observers.find(function (observer) {
      return observer.getCurrentResult().isStale && observer.config.enabled && (observer.config.refetchOnWindowFocus && type === 'focus' || observer.config.refetchOnReconnect && type === 'online');
    });

    if (staleObserver) {
      staleObserver.fetch();
    } // Continue any paused fetch


    this.continue();
  };

  _proto.subscribe = function subscribe(listener) {
    var observer = new QueryObserver(this.config);
    observer.subscribe(listener);
    return observer;
  };

  _proto.subscribeObserver = function subscribeObserver(observer) {
    this.observers.push(observer); // Stop the query from being garbage collected

    this.clearGcTimeout();
  };

  _proto.unsubscribeObserver = function unsubscribeObserver(observer) {
    this.observers = this.observers.filter(function (x) {
      return x !== observer;
    });

    if (!this.observers.length) {
      // If the transport layer does not support cancellation
      // we'll let the query continue so the result can be cached
      if (this.isTransportCancelable) {
        this.cancel();
      }

      this.scheduleGc();
    }
  };

  _proto.refetch = function refetch(options, config) {
    try {
      var _this5 = this;

      return _catch(function () {
        return _await(_this5.fetch(undefined, config));
      }, function (error) {
        if ((options == null ? void 0 : options.throwOnError) === true) {
          throw error;
        }
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetchMore = function fetchMore(fetchMoreVariable, options, config) {
    try {
      var _this7 = this;

      return _this7.fetch({
        fetchMore: {
          fetchMoreVariable: fetchMoreVariable,
          previous: (options == null ? void 0 : options.previous) || false
        }
      }, config);
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetch = function fetch(options, config) {
    try {
      var _exit2 = false;

      var _this9 = this;

      return _invoke(function () {
        if (_this9.promise) {
          return function () {
            if ((options == null ? void 0 : options.fetchMore) && _this9.state.data) {
              // Silently cancel current fetch if the user wants to fetch more
              return _awaitIgnored(_this9.cancel(true));
            } else {
              // Return current promise if we are already fetching
              _exit2 = true;
              return _this9.promise;
            }
          }();
        }
      }, function (_result3) {
        if (_exit2) return _result3;

        // Update config if passed, otherwise the config from the last execution is used
        if (config) {
          _this9.updateConfig(config);
        }

        config = _this9.config; // Get the query function params

        var filter = config.queryFnParamsFilter;
        var params = filter ? filter(_this9.queryKey) : _this9.queryKey;
        _this9.promise = _async(function () {
          return _catch(function () {
            var data;
            return _invoke(function () {
              if (config.infinite) {
                return _await(_this9.startInfiniteFetch(config, params, options), function (_this9$startInfiniteF) {
                  data = _this9$startInfiniteF;
                });
              } else {
                return _await(_this9.startFetch(config, params, options), function (_this9$startFetch) {
                  data = _this9$startFetch;
                });
              }
            }, function () {
              // Set success state
              _this9.setData(data); // Cleanup


              delete _this9.promise; // Return data

              return data;
            });
          }, function (error) {
            // Set error state if needed
            if (!(isCancelledError(error) && error.silent)) {
              _this9.dispatch({
                type: ActionType.Error,
                error: error
              });
            } // Log error


            if (!isCancelledError(error)) {
              Console.error(error);
            } // Cleanup


            delete _this9.promise; // Propagate error

            throw error;
          });
        })();
        return _this9.promise;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.startFetch = function startFetch(config, params, _options) {
    try {
      var _this11 = this;

      // Create function to fetch the data
      var fetchData = function fetchData() {
        return config.queryFn.apply(config, params);
      }; // Set to fetching state if not already in it


      if (!_this11.state.isFetching) {
        _this11.dispatch({
          type: ActionType.Fetch
        });
      } // Try to fetch the data


      return _this11.tryFetchData(config, fetchData);
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.startInfiniteFetch = function startInfiniteFetch(config, params, options) {
    try {
      var _this13 = this;

      var fetchMore = options == null ? void 0 : options.fetchMore;

      var _ref = fetchMore || {},
          previous = _ref.previous,
          fetchMoreVariable = _ref.fetchMoreVariable;

      var isFetchingMore = fetchMore ? previous ? 'previous' : 'next' : false;
      var prevPages = _this13.state.data || []; // Create function to fetch a page

      var fetchPage = _async(function (pages, prepend, cursor) {
        var lastPage = getLastPage(pages, prepend);

        if (typeof cursor === 'undefined' && typeof lastPage !== 'undefined' && config.getFetchMore) {
          cursor = config.getFetchMore(lastPage, pages);
        }

        return _await(config.queryFn.apply(config, params.concat([cursor])), function (page) {
          return prepend ? [page].concat(pages) : [].concat(pages, [page]);
        });
      }); // Create function to fetch the data


      var fetchData = function fetchData() {
        if (isFetchingMore) {
          return fetchPage(prevPages, previous, fetchMoreVariable);
        } else if (!prevPages.length) {
          return fetchPage([]);
        } else {
          var promise = fetchPage([]);

          for (var i = 1; i < prevPages.length; i++) {
            promise = promise.then(fetchPage);
          }

          return promise;
        }
      }; // Set to fetching state if not already in it


      if (!_this13.state.isFetching || _this13.state.isFetchingMore !== isFetchingMore) {
        _this13.dispatch({
          type: ActionType.Fetch,
          isFetchingMore: isFetchingMore
        });
      } // Try to get the data


      return _this13.tryFetchData(config, fetchData);
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.tryFetchData = function tryFetchData(config, fn) {
    try {
      var _this15 = this;

      return new Promise(function (outerResolve, outerReject) {
        var resolved = false;
        var continueLoop;
        var cancelTransport;

        var done = function done() {
          resolved = true;
          delete _this15.cancelFetch;
          delete _this15.continueFetch;
          delete _this15.isTransportCancelable; // End loop if currently paused

          continueLoop == null ? void 0 : continueLoop();
        };

        var resolve = function resolve(value) {
          done();
          outerResolve(value);
        };

        var reject = function reject(value) {
          done();
          outerReject(value);
        }; // Create callback to cancel this fetch


        _this15.cancelFetch = function (silent) {
          reject(new CancelledError(silent));
          cancelTransport == null ? void 0 : cancelTransport();
        }; // Create callback to continue this fetch


        _this15.continueFetch = function () {
          continueLoop == null ? void 0 : continueLoop();
        }; // Create loop function


        var run = _async(function () {
          return _catch(function () {
            // Execute query
            var promiseOrValue = fn(); // Check if the transport layer support cancellation

            if (isCancelable(promiseOrValue)) {
              cancelTransport = function cancelTransport() {
                try {
                  promiseOrValue.cancel();
                } catch (_unused) {}
              };

              _this15.isTransportCancelable = true;
            } // Await data


            return _await(promiseOrValue, function (_promiseOrValue) {
              resolve(_promiseOrValue);
            });
          }, function (error) {
            // Stop if the fetch is already resolved
            if (resolved) {
              return;
            } // Do we need to retry the request?


            var failureCount = _this15.state.failureCount;
            var retry = config.retry,
                retryDelay = config.retryDelay;
            var shouldRetry = retry === true || failureCount < retry || typeof retry === 'function' && retry(failureCount, error);

            if (!shouldRetry) {
              // We are done if the query does not need to be retried
              reject(error);
              return;
            } // Increase the failureCount


            _this15.dispatch({
              type: ActionType.Failed
            }); // Delay


            return _await(sleep(functionalUpdate(retryDelay, failureCount) || 0), function () {
              // Pause retry if the document is not visible or when the device is offline
              return _invoke(function () {
                if (!isDocumentVisible() || !isOnline()) {
                  return _awaitIgnored(new Promise(function (continueResolve) {
                    continueLoop = continueResolve;
                  }));
                }
              }, function () {
                if (!resolved) {
                  run();
                }
              }); // Try again if not resolved yet
            });
          });
        }); // Start loop


        run();
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  return Query;
}();

function getLastPage(pages, previous) {
  return previous ? pages[0] : pages[pages.length - 1];
}

function hasMorePages(config, pages, previous) {
  if (config.infinite && config.getFetchMore && Array.isArray(pages)) {
    return Boolean(config.getFetchMore(getLastPage(pages, previous), pages));
  }
}

function getDefaultState(config) {
  var initialData = typeof config.initialData === 'function' ? config.initialData() : config.initialData;
  var hasInitialData = typeof initialData !== 'undefined';
  var initialStatus = hasInitialData ? QueryStatus.Success : config.enabled ? QueryStatus.Loading : QueryStatus.Idle;
  return _extends({}, getStatusProps(initialStatus), {
    canFetchMore: hasMorePages(config, initialData),
    data: initialData,
    error: null,
    failureCount: 0,
    isFetching: initialStatus === QueryStatus.Loading,
    isFetchingMore: false,
    isInitialData: true,
    updateCount: 0,
    updatedAt: Date.now()
  });
}

export function queryReducer(state, action) {
  var _action$updatedAt;

  switch (action.type) {
    case ActionType.Failed:
      return _extends({}, state, {
        failureCount: state.failureCount + 1
      });

    case ActionType.Fetch:
      var status = typeof state.data !== 'undefined' ? QueryStatus.Success : QueryStatus.Loading;
      return _extends({}, state, getStatusProps(status), {
        isFetching: true,
        isFetchingMore: action.isFetchingMore || false,
        failureCount: 0
      });

    case ActionType.Success:
      return _extends({}, state, getStatusProps(QueryStatus.Success), {
        canFetchMore: action.canFetchMore,
        data: action.data,
        error: null,
        failureCount: 0,
        isFetching: false,
        isFetchingMore: false,
        isInitialData: false,
        updateCount: state.updateCount + 1,
        updatedAt: (_action$updatedAt = action.updatedAt) != null ? _action$updatedAt : Date.now()
      });

    case ActionType.Error:
      return _extends({}, state, getStatusProps(QueryStatus.Error), {
        error: action.error,
        failureCount: state.failureCount + 1,
        isFetching: false,
        isFetchingMore: false,
        throwInErrorBoundary: true,
        updateCount: state.updateCount + 1
      });

    default:
      return state;
  }
}