"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.QueryObserver = void 0;

var _extends2 = _interopRequireDefault(require("@babel/runtime/helpers/extends"));

var _utils = require("./utils");

function _await(value, then, direct) {
  if (direct) {
    return then ? then(value) : value;
  }

  if (!value || !value.then) {
    value = Promise.resolve(value);
  }

  return then ? value.then(then) : value;
}

function _empty() {}

function _catch(body, recover) {
  try {
    var result = body();
  } catch (e) {
    return recover(e);
  }

  if (result && result.then) {
    return result.then(void 0, recover);
  }

  return result;
}

var QueryObserver = /*#__PURE__*/function () {
  function QueryObserver(config) {
    this.config = config;
    this.initialUpdateCount = 0; // Bind exposed methods

    this.clear = this.clear.bind(this);
    this.refetch = this.refetch.bind(this);
    this.fetchMore = this.fetchMore.bind(this); // Subscribe to the query

    this.updateQuery();
  }

  var _proto = QueryObserver.prototype;

  _proto.subscribe = function subscribe(listener) {
    this.listener = listener;
    this.currentQuery.subscribeObserver(this);

    if (this.config.enabled && this.config.forceFetchOnMount) {
      this.fetch();
    } else {
      this.optionalFetch();
    }

    this.updateTimers();
    return this.unsubscribe.bind(this);
  };

  _proto.unsubscribe = function unsubscribe() {
    this.listener = undefined;
    this.clearTimers();
    this.currentQuery.unsubscribeObserver(this);
  };

  _proto.updateConfig = function updateConfig(config) {
    var prevConfig = this.config;
    var prevQuery = this.currentQuery;
    this.config = config;
    this.updateQuery(); // Take no further actions if there is no subscriber

    if (!this.listener) {
      return;
    } // If we subscribed to a new query, optionally fetch and update refetch


    if (this.currentQuery !== prevQuery) {
      this.optionalFetch();
      this.updateTimers();
      return;
    } // Optionally fetch if the query became enabled


    if (config.enabled && !prevConfig.enabled) {
      this.optionalFetch();
    } // Update stale interval if needed


    if (config.enabled !== prevConfig.enabled || config.staleTime !== prevConfig.staleTime) {
      this.updateStaleTimeout();
    } // Update refetch interval if needed


    if (config.enabled !== prevConfig.enabled || config.refetchInterval !== prevConfig.refetchInterval) {
      this.updateRefetchInterval();
    }
  };

  _proto.getCurrentQuery = function getCurrentQuery() {
    return this.currentQuery;
  };

  _proto.getCurrentResult = function getCurrentResult() {
    return this.currentResult;
  };

  _proto.clear = function clear() {
    return this.currentQuery.clear();
  };

  _proto.refetch = function refetch(options) {
    try {
      var _this2 = this;

      return _this2.currentQuery.refetch(options, _this2.config);
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetchMore = function fetchMore(fetchMoreVariable, options) {
    try {
      var _this4 = this;

      return _this4.currentQuery.fetchMore(fetchMoreVariable, options, _this4.config);
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetch = function fetch() {
    var _this5 = this;

    return _await(_catch(function () {
      return _await(_this5.currentQuery.fetch(undefined, _this5.config));
    }, _empty));
  };

  _proto.optionalFetch = function optionalFetch() {
    if (this.config.enabled && // Don't auto refetch if disabled
    !(this.config.suspense && this.currentResult.isFetched) && // Don't refetch if in suspense mode and the data is already fetched
    this.currentResult.isStale && ( // Only refetch if stale
    this.config.refetchOnMount || this.currentQuery.observers.length === 1)) {
      this.fetch();
    }
  };

  _proto.notify = function notify() {
    var _this$listener;

    (_this$listener = this.listener) == null ? void 0 : _this$listener.call(this, this.currentResult);
  };

  _proto.updateStaleTimeout = function updateStaleTimeout() {
    var _this6 = this;

    if (_utils.isServer) {
      return;
    }

    this.clearStaleTimeout();

    if (this.currentResult.isStale || !(0, _utils.isValidTimeout)(this.config.staleTime)) {
      return;
    }

    var timeElapsed = Date.now() - this.currentResult.updatedAt;
    var timeUntilStale = this.config.staleTime - timeElapsed + 1;
    var timeout = Math.max(timeUntilStale, 0);
    this.staleTimeoutId = setTimeout(function () {
      if (!_this6.currentResult.isStale) {
        _this6.currentResult = (0, _extends2.default)({}, _this6.currentResult, {
          isStale: true
        });

        _this6.notify();

        _this6.config.queryCache.notifyGlobalListeners(_this6.currentQuery);
      }
    }, timeout);
  };

  _proto.updateRefetchInterval = function updateRefetchInterval() {
    var _this7 = this;

    if (_utils.isServer) {
      return;
    }

    this.clearRefetchInterval();

    if (!this.config.enabled || !(0, _utils.isValidTimeout)(this.config.refetchInterval)) {
      return;
    }

    this.refetchIntervalId = setInterval(function () {
      if (_this7.config.refetchIntervalInBackground || (0, _utils.isDocumentVisible)()) {
        _this7.fetch();
      }
    }, this.config.refetchInterval);
  };

  _proto.updateTimers = function updateTimers() {
    this.updateStaleTimeout();
    this.updateRefetchInterval();
  };

  _proto.clearTimers = function clearTimers() {
    this.clearStaleTimeout();
    this.clearRefetchInterval();
  };

  _proto.clearStaleTimeout = function clearStaleTimeout() {
    if (this.staleTimeoutId) {
      clearInterval(this.staleTimeoutId);
      this.staleTimeoutId = undefined;
    }
  };

  _proto.clearRefetchInterval = function clearRefetchInterval() {
    if (this.refetchIntervalId) {
      clearInterval(this.refetchIntervalId);
      this.refetchIntervalId = undefined;
    }
  };

  _proto.updateResult = function updateResult() {
    var currentQuery = this.currentQuery,
        currentResult = this.currentResult,
        previousQueryResult = this.previousQueryResult,
        config = this.config;
    var state = currentQuery.state;
    var data = state.data,
        status = state.status,
        updatedAt = state.updatedAt;
    var isPreviousData = false; // Keep previous data if needed

    if (config.keepPreviousData && (state.isIdle || state.isLoading) && (previousQueryResult == null ? void 0 : previousQueryResult.isSuccess)) {
      data = previousQueryResult.data;
      updatedAt = previousQueryResult.updatedAt;
      status = previousQueryResult.status;
      isPreviousData = true;
    }

    var isStale; // When the query has not been fetched yet and this is the initial render,
    // determine the staleness based on the initialStale or existence of initial data.

    if (!currentResult && state.isInitialData) {
      if (typeof config.initialStale === 'function') {
        isStale = config.initialStale();
      } else if (typeof config.initialStale === 'boolean') {
        isStale = config.initialStale;
      } else {
        isStale = typeof state.data === 'undefined';
      }
    } else {
      isStale = currentQuery.isStaleByTime(config.staleTime);
    }

    this.currentResult = (0, _extends2.default)({}, (0, _utils.getStatusProps)(status), {
      canFetchMore: state.canFetchMore,
      clear: this.clear,
      data: data,
      error: state.error,
      failureCount: state.failureCount,
      fetchMore: this.fetchMore,
      isFetched: state.updateCount > 0,
      isFetchedAfterMount: state.updateCount > this.initialUpdateCount,
      isFetching: state.isFetching,
      isFetchingMore: state.isFetchingMore,
      isInitialData: state.isInitialData,
      isPreviousData: isPreviousData,
      isStale: isStale,
      refetch: this.refetch,
      updatedAt: updatedAt
    });
  };

  _proto.updateQuery = function updateQuery() {
    var prevQuery = this.currentQuery; // Remove the initial data when there is an existing query
    // because this data should not be used for a new query

    var config = this.config.keepPreviousData && prevQuery ? (0, _extends2.default)({}, this.config, {
      initialData: undefined
    }) : this.config;
    var query = config.queryCache.getQueryByHash(config.queryHash);

    if (!query) {
      query = config.queryCache.createQuery(config);
    }

    if (query === prevQuery) {
      return;
    }

    this.previousQueryResult = this.currentResult;
    this.currentQuery = query;
    this.initialUpdateCount = query.state.updateCount;
    this.updateResult();

    if (this.listener) {
      prevQuery == null ? void 0 : prevQuery.unsubscribeObserver(this);
      this.currentQuery.subscribeObserver(this);
    }
  };

  _proto.onQueryUpdate = function onQueryUpdate(action) {
    var type = action.type; // Store current result and get new result

    var prevResult = this.currentResult;
    this.updateResult();
    var currentResult = this.currentResult,
        config = this.config; // We need to check the action because the state could have
    // transitioned from success to success in case of `setQueryData`.

    if (type === 2) {
      config.onSuccess == null ? void 0 : config.onSuccess(currentResult.data);
      config.onSettled == null ? void 0 : config.onSettled(currentResult.data, null);
      this.updateTimers();
    } else if (type === 3) {
      config.onError == null ? void 0 : config.onError(currentResult.error);
      config.onSettled == null ? void 0 : config.onSettled(undefined, currentResult.error);
      this.updateTimers();
    }

    if ( // Always notify on data or error change
    currentResult.data !== prevResult.data || currentResult.error !== prevResult.error || // Maybe notify on other changes
    config.notifyOnStatusChange) {
      this.notify();
    }
  };

  return QueryObserver;
}();

exports.QueryObserver = QueryObserver;